(* Programming and Reasoning with Infinite Data in Isabelle/HOL (POPL 2018) *)
(* Mathias Fleury, Andreas Lochbihler, Andrei Popescu *)

theory Recursive_Types imports
  Main
  "~~/src/HOL/Library/BNF_Corec"
  "~~/src/HOL/Library/Simps_Case_Conv"
begin

hide_type (open) type

type_synonym var = nat

(* syntactic domain *)
datatype ty 
  = Var var
  | Arrow ty ty
  | Mu var ty

(* semantic domain *)
codatatype type
  = is_TVar: TVar var
  | TArrow type type (infixr "\<rightarrow>" 120)



corec tsubst :: "var \<Rightarrow> type \<Rightarrow> type \<Rightarrow> type" ("[_/ \<mapsto>/ _]\<cdot>_" [0, 0, 130] 130) where 
  "[X \<mapsto> t1]\<cdot>t2 = 
  (case t2 of TVar Y \<Rightarrow> if X = Y then t1 else TVar Y
   |        t2' \<rightarrow> t2'' \<Rightarrow> [X \<mapsto> t1]\<cdot>t2' \<rightarrow> [X \<mapsto> t1]\<cdot>t2'')"

simps_of_case tsubst_simps [simp]: tsubst.code

friend_of_corec tsubst where
  -- \<open>Currently, for friendship we need to make guardedness explicit. (**) \<close>
  "[X \<mapsto> t1]\<cdot>t2 = (case t2 of TVar Y \<Rightarrow> 
     if X = Y then (case t1 of TVar Z \<Rightarrow> TVar Z | t1' \<rightarrow> t1'' \<Rightarrow> t1' \<rightarrow> t1'') else TVar Y
   | t2' \<rightarrow> t2'' \<Rightarrow> [X \<mapsto> t1]\<cdot>t2' \<rightarrow> [X \<mapsto> t1]\<cdot>t2'')"
  subgoal by(simp split: type.split)
  subgoal by transfer_prover
  done

text \<open>
  The fixpoint operator on tree types. The unguarded case of \<open>t = Var X\<close> is set to \<open>Var X\<close>
  to avoid pointless case distinctions.
\<close> 

corec mu :: "var \<Rightarrow> type \<Rightarrow> type" ("\<mu> _./ _" [0, 110] 109) where
  "\<mu> X. t =
  (case t of TVar Y \<Rightarrow> if X = Y then undefined else TVar Y
   |       t1 \<rightarrow> t2 \<Rightarrow> [X \<mapsto> \<mu> X. t]\<cdot>t1 \<rightarrow> [X \<mapsto> \<mu> X. t]\<cdot>t2)"

lemma mu_TVar [simp]: "\<mu> X. TVar Y = TVar Y" if "X \<noteq> Y"
  by(simp add: mu.code that)

lemma mu_unfold: "\<mu> X. t = [X \<mapsto> \<mu> X. t]\<cdot>t" if "t \<noteq> TVar X"
   by(subst mu.code)(simp add: that split: type.split if_split_asm)



fun sem :: "ty \<Rightarrow> type" ("\<lbrakk>_\<rbrakk>") where
  "\<lbrakk>Var x\<rbrakk> = TVar x"
| "\<lbrakk>Arrow T1 T2\<rbrakk> = \<lbrakk>T1\<rbrakk> \<rightarrow> \<lbrakk>T2\<rbrakk>"
| "\<lbrakk>Mu X T\<rbrakk> = \<mu> X. \<lbrakk>T\<rbrakk>"


(* All occurrences of X in T are guarded by an arrow. *)
fun contractive :: "var \<Rightarrow> ty \<Rightarrow> bool" where
  "contractive X (Var Y) = (X \<noteq> Y)"
| "contractive X (Arrow T T') = True"
| "contractive X (Mu Y T) = contractive X T"

fun guarded :: "ty \<Rightarrow> bool" where
  "guarded (Var X) = True"
| "guarded (Arrow T T') = (guarded T \<and> guarded T')"
| "guarded (Mu X T) = (guarded T \<and> contractive X T)"

lemma contractive_correct:
  "contractive X T = (\<lbrakk>T\<rbrakk> \<noteq> TVar X)" if "guarded T"
  using that
  apply(induction T arbitrary: X)
  subgoal Var by auto
  subgoal Arrow for T T' by simp
  subgoal Mu for Y T X
    apply(clarsimp)
    apply(cases T)
    subgoal Var by simp
    subgoal by(subst mu_unfold) simp_all
    subgoal for Z T by(cases "\<mu> Z. \<lbrakk>T\<rbrakk>", simp)(subst mu_unfold, simp_all)
    done
  done

lemma Mu_unfold: "\<lbrakk>Mu X T\<rbrakk> = [X \<mapsto> \<lbrakk>Mu X T\<rbrakk>]\<cdot>\<lbrakk>T\<rbrakk>" if "contractive X T" "guarded T"
  using that unfolding sem.simps
  by(subst mu_unfold)(simp_all add: contractive_correct)

end