(* Programming and Reasoning with Infinite Data in Isabelle/HOL (POPL 2018) *)
(* Mathias Fleury, Andreas Lochbihler, Andrei Popescu *)

theory Stream_Processor imports
  "~~/src/HOL/Library/Stream"
  "~~/src/HOL/Library/BNF_Corec"
  "~~/src/HOL/Library/Code_Test"
begin

section {* Stream Processors *}

syntax -- \<open>Pretty syntax for streams\<close>
  "_stream" :: "args => 'a list"    ("[(_), ....]")
translations
  "[x, xs, ....]" == "CONST SCons x [xs, ....]"
  "[x, ....]" == "CONST SCons x (CONST undefined)"


datatype (discs_sels) ('in, 'out, 'c) sp\<^sub>\<mu> 
  = Get \<open>'in \<Rightarrow> ('in, 'out, 'c) sp\<^sub>\<mu>\<close> 
  | Put \<open>'out\<close> \<open>'c\<close>
codatatype ('in, 'out) sp\<^sub>\<nu> =
  In (out: \<open>('in, 'out, ('in, 'out) sp\<^sub>\<nu>) sp\<^sub>\<mu>\<close>)

declare sp\<^sub>\<mu>.map_transfer[transfer_rule]


text \<open> We define the direct subterm relation for \<open>sp\<^sub>\<mu>\<close> and prove it well-founded (predicate \<open>wf\<close>),
  which we will need to establish termination of the recursion in the corecursive definitions below.\<close>
definition "sub \<equiv> {(f a, Get f) | a f. True}"

lemma subI [intro]: "(f a, Get f) \<in> sub"
  unfolding sub_def by blast

lemma wf_sub [simp, intro]: "wf sub"
proof (rule wfUNIVI)
  fix P  :: "('a, 'b, 'c) sp\<^sub>\<mu> \<Rightarrow> bool" and x
  assume "\<forall>x. (\<forall>y. (y, x) \<in> sub \<longrightarrow> P y) \<longrightarrow> P x"
  hence I: "\<And>x. (\<forall>y. (\<exists>a f. y = f a \<and> x = Get f) \<longrightarrow> P y) \<Longrightarrow> P x" unfolding sub_def by blast
  show "P x" by (induct x) (auto intro: I)
qed

text \<open>Running a stream processor on a stream yields another stream\<close>

corecursive run :: "('a, 'b) sp\<^sub>\<nu> \<Rightarrow> 'a stream \<Rightarrow> 'b stream" where
  "run sp s = (case out sp of
     Get f \<Rightarrow> run (In (f (shd s))) (stl s)
   | Put b sp \<Rightarrow> b ## run sp s)"
  by(relation "map_prod In In ` sub <*lex*> {}")
    (auto simp: inj_on_def out_def split: sp\<^sub>\<nu>.splits intro: wf_map_prod_image)

corec copy :: "('a, 'a) sp\<^sub>\<nu>" where
  "copy = In (Get (\<lambda>a. Put a copy))"

corec (friend) get :: "('a \<Rightarrow> ('a, 'b) sp\<^sub>\<nu>) \<Rightarrow> ('a, 'b) sp\<^sub>\<nu>" where
  "get f = In (Get (\<lambda>a. out (f a)))"

abbreviation put :: "'b \<Rightarrow> ('a, 'b) sp\<^sub>\<nu> \<Rightarrow> ('a, 'b) sp\<^sub>\<nu>" where
  "put x c \<equiv> In (Put x c)"

lemma run_simps [simp]:
  "run (In (Get f)) s = run (In (f (shd s))) (stl s)"
  "run (In (Put b sp)) s = b ## run sp s"
  by(subst run.code; simp; fail)+

lemma copy_sel [simp]: "out copy = Get (\<lambda>a. Put a copy)"
  by (subst copy.code; simp)

text \<open>Composition of stream processors\<close>

corecursive sp_comp :: "('a, 'b) sp\<^sub>\<nu> \<Rightarrow> ('c, 'a) sp\<^sub>\<nu> \<Rightarrow> ('c, 'b) sp\<^sub>\<nu>" (infixl "oo" 65) where
  "sp oo sp' = (case (out sp, out sp') of
      (Put b sp, _) \<Rightarrow> In (Put b (sp oo sp'))
    | (Get f, Put b sp) \<Rightarrow> In (f b) oo sp
    | (_, Get g) \<Rightarrow> get (\<lambda>a. (sp oo In (g a))))"
  by(relation "map_prod In In ` sub <*lex*> map_prod In In ` sub")
    (auto simp: inj_on_def out_def split: sp\<^sub>\<nu>.splits intro: wf_map_prod_image)


subsection \<open>Example: Calculator\<close>

text \<open>Inspired by Hur et al.: The power of parameterization in coinductive proof. POPL 2013, 193--206.\<close>

type_synonym calc = "nat \<Rightarrow> (nat, nat) sp\<^sub>\<nu>"

corec (friend) restart :: "calc \<Rightarrow> calc" where
  "restart f n = (if n > 0 then put n (restart f (n - 1)) else In (out (f 0)))"

lemma restart_simps: "restart f n = (if n > 0 then put n (restart f (n - 1)) else f 0)"
  by(subst restart.code) simp

lemma restart_cong: "restart f n = restart g n" if "f 0 = g 0"
  using that by(induction n)(simp_all add: restart.code)

corec f :: "calc" where
  "f n = put n (get (\<lambda>v. if v \<noteq> 0 then f (2 * v + n) else restart f (v + n)))"

lemma f_sel [simp]: "out (f n) = Put n (get (\<lambda>v. if v \<noteq> 0 then f (2 * v + n) else restart f (v + n)))"
  by(simp add: f.code)

corec g :: "calc" where
  "g m = put (2 * m) (get (\<lambda>v. if v = 0 then restart g (2 * m) else g (v + m)))"

lemma g_sel [simp]: "out (g m) = Put (2 * m) (get (\<lambda>v. if v = 0 then restart g (2 * m) else g (v + m)))"
  by(simp add: g.code)

lemma g_sim_f: "g m = f (2 * m)"
proof(coinduction arbitrary: m rule: restart.coinduct)
  case (Eq_nat_nat_sp\<^sub>\<nu> m)
  have "restart g (m * 2) = restart (\<lambda>_. g 0) (m * 2)"
    and "restart f (m * 2) = restart (\<lambda>_. f 0) (m * 2)" by(rule restart_cong; simp)+
  then show ?case
    by(auto 4 3 intro!: sp\<^sub>\<nu>.cong_get sp\<^sub>\<nu>.cong_restart rel_funI exI simp add: mult.commute add_mult_distrib intro: sp\<^sub>\<nu>.cong_base)
qed

export_code restart f g in Haskell module_name Calculator

value [GHC] "stake 15 (run (g 0) [1, 2, 0, 3, 0, ....])"
value [GHC] "stake 69 (run (g 0 oo f 0) [1, 2, 0, ....])"

end
