(* Programming and Reasoning with Infinite Data in Isabelle/HOL (POPL 2018) *)
(* Mathias Fleury, Andreas Lochbihler, Andrei Popescu *)

theory Tree imports
  "~~/src/HOL/Library/Stream"
  "~~/src/HOL/Library/BNF_Corec"
  "~~/src/HOL/Library/Simps_Case_Conv"
  "~~/src/HOL/Library/Code_Test"
begin

friend_of_corec smap :: "(int \<Rightarrow> int) \<Rightarrow> int stream \<Rightarrow> int stream" where
  "smap f s = (case s of x ## s' \<Rightarrow> f x ## smap f s')"
  subgoal by(simp split: stream.split)
  subgoal by transfer_prover
  done

corec up :: "int stream" where
  "up = 0 ## smap (\<lambda>x. x + 1) up"

value [GHC] "stake 10 up"

codatatype 'a tree 
  = Node (left: \<open>'a tree\<close>) (root: 'a) (right: \<open>'a tree\<close>)

corec (friend) tmap :: "('a \<Rightarrow> 'a) \<Rightarrow> 'a tree \<Rightarrow> 'a tree" where
  "tmap f t = (case t of Node l x r \<Rightarrow> 
   Node (tmap f l) (f x) (tmap f r))"

(* Tree of all numbers *)

corec tnum :: "int tree" where
  "tnum = Node (tmap (\<lambda>x. 2 * x + 1) tnum) 0 (tmap (\<lambda>x. 2 * x + 2) tnum)"

corec chop :: \<open>'a tree \<Rightarrow> 'a tree\<close> where
  "chop t = (case t of Node l _ r \<Rightarrow> Node r (root l) (chop l))"

corec stream_of :: \<open>'a tree \<Rightarrow> 'a stream\<close> where
  "stream_of t = root t ## stream_of (chop t)"

value [GHC] "stake 10 (stream_of tnum)"

(* Some useful lemmas *)

lemma tnum_sel [simp]:
  "root tnum = 0"
  "left tnum = tmap (\<lambda>x. 2 * x + 1) tnum"
  "right tnum = tmap (\<lambda>x. 2 * x + 2) tnum"
  by(subst tnum.code; simp; fail)+

simps_of_case tmap_simps [simp]: tmap.code
thm tmap_simps

lemma tmap_sel [simp]:
  "root (tmap f t) = f (root t)"
  "left (tmap f t) = tmap f (left t)"
  "right (tmap f t) = tmap f (right t)"
  by(cases t; simp; fail)+

lemma tmap_comp: "tmap f (tmap g t) = tmap (f \<circ> g) t"
  by(coinduction arbitrary: t) auto

lemma stream_of_sel [simp]:
  "shd (stream_of t) = root t"
  "stl (stream_of t) = stream_of (chop t)"
  by(subst stream_of.code; simp; fail)+

simps_of_case chop_simps [simp]: chop.code

lemma chop_sel [simp]:
  "root (chop t) = root (left t)"
  "left (chop t) = right t"
  "right (chop t) = chop (left t)"
  by(cases t; simp; fail)+

lemma chop_tmap: "chop (tmap f t) = tmap f (chop t)"
  by(coinduction arbitrary: t rule: tree.coinduct_strong) auto

lemma stream_of_tmap [simp]: "stream_of (tmap f t) = smap f (stream_of t)"
  by(coinduction arbitrary: t)(auto simp add: chop_tmap)


lemma chop_tnum: "chop tnum = tmap (\<lambda>x. x + 1) tnum"
proof(coinduction rule: tnum.coinduct)
  case Eq_int_tree
  let ?R = "\<lambda>t t'. t = chop tnum \<and> t' = tmap (\<lambda>x. x + 1) tnum"
  have "tree.congclp ?R (left (chop tnum)) (left (tmap (\<lambda>x. x + 1) tnum))"
    by(simp add: tmap_comp o_def algebra_simps tmap.corec.cong_refl)
  moreover have "root (chop tnum) = root (tmap (\<lambda>x. x + 1) tnum)" by simp
  moreover
  have "right (tmap (\<lambda>x. x + 1) tnum) = tmap (\<lambda>x. 2 * x + 1) (tmap (\<lambda>x. x + 1) tnum)"
    by(simp add: tmap_comp o_def)
  then have "tree.congclp ?R (right (chop tnum)) (right (tmap (\<lambda>x. x + 1) tnum))"
    by(simp add: chop_tmap tmap.corec.cong_tmap[OF refl] tmap.corec.cong_base)
  ultimately show ?case by blast
qed


lemma "stream_of tnum = up"
  apply(rule up.unique) (* Proof by uniqueness *)
  apply(subst stream_of.code)
  apply(simp add: chop_tnum)
  done

end