(* Programming and Reasoning with Infinite Data in Isabelle/HOL (POPL 2018) *)
(* Mathias Fleury, Andreas Lochbihler, Andrei Popescu *)

theory While_Language imports
  Main
  "~~/src/HOL/Library/BNF_Corec"
  "~~/src/HOL/Library/Simps_Case_Conv"
begin

declare if_weak_cong[cong del]

text \<open>Declare some types rather than use polymorphism, 
 to keep notations short.\<close>

typedecl state


section \<open>Semantic processes\<close>
 
text \<open>Steps are either state updates or test results.\<close>
type_synonym step = "state + bool"
abbreviation Update :: "state \<Rightarrow> step" where "Update \<equiv> Inl"
abbreviation Branch :: "bool \<Rightarrow> step" where "Branch \<equiv> Inr"

text \<open>A trace is a possibly infinite sequence of steps and,
  if the sequence is finite, a final state. A process is a 
  function from states to traces.
\<close>
codatatype trace 
  = is_Final: Final (final: state) 
  | Step (step: "step") (rest: "trace")

text \<open>Pretty syntax for traces\<close>
syntax "_trace" :: "args \<Rightarrow> state \<Rightarrow> trace" ("[(_)/ \<bar> _]")
translations
  "[x, xs \<bar> s]" == "CONST Step x [xs \<bar> s]"
  "[x \<bar> s]" == "CONST Step x (CONST Final s)"
notation Final ("[ \<bar> _]")

type_synonym process = "state \<Rightarrow> trace"


definition skip :: "process" where "skip s = [ \<bar> s]"

(* Atomic processes are parameterized by semantic actions, atm :: state \<Rightarrow> state *)
definition atom :: "(state \<Rightarrow> state) \<Rightarrow> process" where 
  "atom atm s = [ Update s \<bar> atm s ]"

corec prepend :: "trace \<Rightarrow> process \<Rightarrow> trace" where
  "prepend tr p = 
  (case tr of Final s \<Rightarrow> p s
     | Step x tr' \<Rightarrow> Step x (prepend tr' p))"

definition seq :: "process \<Rightarrow> process \<Rightarrow> process" where 
  "seq p1 p2 s = prepend (p1 s) p2"

simps_of_case prepend_simps [simp]: prepend.code

friend_of_corec prepend where
  "prepend tr p =
  (case tr of Final s \<Rightarrow> (case p s of Final s' \<Rightarrow> Final s' | Step x tr' \<Rightarrow> Step x tr')
     | Step x tr' \<Rightarrow> Step x (prepend tr' p))"
  subgoal by(simp split: trace.split)
  subgoal by transfer_prover
  done


(* If and while are parameterized by semantic tests, tst :: state \<Rightarrow> bool *)
definition conditional :: "(state \<Rightarrow> bool) \<Rightarrow> process \<Rightarrow> process \<Rightarrow> process" where 
  "conditional tst p1 p2 s = 
   (let b = tst s in Step (Branch b) (if b then p1 s else p2 s))"

corec (friend) while :: "(state \<Rightarrow> bool) \<Rightarrow> process \<Rightarrow> process" where 
  "while tst p s = 
   (let b = tst s in 
    Step (Branch b) (if b then prepend (p s) (while tst p) else Final s))"

theorem while_eq: 
  "while tst p = conditional tst (seq p (while tst p)) skip"
  apply(rule ext)
  apply(subst while.code)
  apply(simp add: conditional_def seq_def skip_def)
  done


section \<open>Command syntax and trace function\<close>

typedecl atom
typedecl test

datatype com
  = SKIP
  | ATOM atom
  | SEQ com com (infixr ";;" 120)
  | IF test com com
  | WHILE test com

(* The trace function is parameterized by evaluation functions for atomic commands and tests: *)
consts evalA :: "atom \<Rightarrow> state \<Rightarrow> state"
consts evalT :: "test \<Rightarrow> state \<Rightarrow> bool"

fun sem :: "com \<Rightarrow> process" ("\<lbrakk>_\<rbrakk>") where 
  "\<lbrakk>SKIP\<rbrakk> = skip"
| "\<lbrakk>ATOM atm\<rbrakk> = atom (evalA atm)"
| "\<lbrakk>c1;; c2\<rbrakk> = seq \<lbrakk>c1\<rbrakk> \<lbrakk>c2\<rbrakk>"
| "\<lbrakk>IF tst c1 c2\<rbrakk> = conditional (evalT tst) \<lbrakk>c1\<rbrakk> \<lbrakk>c2\<rbrakk>"
| "\<lbrakk>WHILE tst c\<rbrakk> = while (evalT tst) \<lbrakk>c\<rbrakk>"

theorem "\<lbrakk>WHILE tst c\<rbrakk> = \<lbrakk>IF tst (SEQ c (WHILE tst c)) SKIP\<rbrakk>"
  unfolding sem.simps
  using while_eq by blast

lemma prepend_skip [simp]: "prepend tr skip = tr"
  apply(coinduction arbitrary: tr)
  subgoal for tr by(cases tr)(simp_all add: skip_def)
  done

theorem "\<lbrakk>SKIP ;; c\<rbrakk> = \<lbrakk>c\<rbrakk>" "\<lbrakk>c ;; SKIP\<rbrakk> = \<lbrakk>c\<rbrakk>"
  by(simp_all add: seq_def skip_def fun_eq_iff)


section \<open>Looping processes\<close>

abbreviation loop :: "process \<Rightarrow> process" where
  "loop \<equiv> while (\<lambda>_. True)"

inductive terminating :: "trace \<Rightarrow> bool" where
  "terminating (Final s)"
| "terminating (Step x tr)" if "terminating tr"

inductive_simps terminating_simps [simp]:
  "terminating (Final s)"
  "terminating (Step x tr)"

proposition nonterminating_prepend [simp]:
  assumes "\<not> terminating tr"
  shows "prepend tr p = tr"
  using assms
  apply(coinduction arbitrary: tr)
  subgoal for tr by(cases tr)(auto)
  done

lemma Final_neq_loop [simp]: "Final s' \<noteq> loop p s"
  by(subst while_eq)(simp add: conditional_def)

proposition nonterminating_loop [simp]: "\<not> terminating (loop p s)"
proof
  assume *: "terminating (loop p s)"
  text \<open>We generalise the trace to a terminating prefix prepended to the loop 
    such that the statement becomes inductive. This generalisation is necessary
    as \<open>terminating\<close>'s inductive definition does not know about the friend \<open>prepend\<close>.\<close>
  define tr where "tr = Final s"
  with * have "terminating (prepend tr (loop p))" by simp
  then show False
    apply(induction "prepend tr (loop p)" arbitrary: tr)
    subgoal Final for s tr by(cases tr) simp_all
    subgoal Step for tr s tr'
      by(cases tr')(subst (asm) (2) while_eq, auto simp add: conditional_def seq_def)
    done
qed

text \<open>Everything after a loop can be discarded.\<close>

theorem seq_loop: "seq (loop p) p' = loop p"
  by(simp add: seq_def fun_eq_iff)

text \<open>Looping is idempotent (except for the first test)\<close>

definition branch_True :: "process \<Rightarrow> process" where
  "branch_True p s = Step (Branch True) (p s)"

theorem loop_loop: "loop (loop p) = branch_True (loop p)"
  by(subst while_eq)(simp add: fun_eq_iff conditional_def seq_loop branch_True_def)

end